import asyncio
import fnmatch
from . import squid


@asyncio.coroutine
def test(request, metadata):
    """Return an ACL action (OK, ERR, or BH) by comparing ACL entries against host metadata"""

    if request.client is None:
        return 'BH', {'log': 'Failed to parse client IP address'}
    if metadata is None or 'instance_id' not in metadata:
        return 'ERR', {'log': 'Metadata not available for this client'}
    else:
        for entry in request.acl:
            if check_acl_entry(entry, metadata):
                return 'OK', {'user': metadata['instance_id']} 
        return 'ERR', {'user': metadata['instance_id']}


def check_acl_entry(entry, metadata):
    """ Check an individual ACL entry against host metadata

    Returns True if matched, else False

    Supported ACL entry strings:
        * Instance ID (i-xxx)
        * Security Group ID (sg-xxx)
        * Image AMI ID (ami-xxx)
        * VPC ID (vpc-xxx)
        * Subnet ID (subnet-xxx)
        * Availability zone (az:us-west-2*)          - Matches shell-style globs
        * Security Group Name (sg:my security group) - Matches shell-style globs
        * Tag (tag:Name=Value)                       - Matches shell-style globs
        * Existence as an EC2 instance (any)         - Matches if request is from a known EC2 instance
    """

    if entry.startswith('i-'):
        return entry == metadata.get('instance_id')

    elif entry.startswith('sg-'):
        for interface in metadata.get('network_interfaces', []):
            for group in interface.get('groups', []):
                if entry == group.get('group_id'):
                    return True
        return False

    elif entry.startswith('ami-'):
        return entry == metadata.get('image_id')

    elif entry.startswith('vpc-'):
        return entry == metadata.get('vpc_id')

    elif entry.startswith('subnet-'):
        for interface in metadata.get('network_interfaces', []):
            if entry == interface.get('subnet_id'):
                return True
        return False

    elif entry.startswith('az:'):
        pattern = entry[3:].lower()
        return fnmatch.fnmatch(metadata.get('placement', {}).get('availability_zone', '').lower(), pattern)

    elif entry.startswith('sg:'):
        pattern = entry[3:].lower()
        for interface in metadata.get('network_interfaces', []):
            for group in interface.get('groups', []):
                if fnmatch.fnmatch(group.get('group_name', '').lower(), pattern):
                    return True
        return False

    elif entry.startswith('tag:'):
        pattern = entry[4:]
        key, pattern = pattern.split('=', 1)
        pattern = pattern.lower()
        if key in metadata.get('tags', {}):
            if fnmatch.fnmatch(metadata['tags'][key].lower(), pattern):
                return True
        return False

    elif entry == 'any':
        return True

    else:
        return False



import ipaddress
from urllib.parse import quote, unquote

# Valid response keywords
_squid_keywords = frozenset(['clt_conn_tag', 'group', 'log', 'message', 'password', 'tag', 'ttl', 'user'])

class Request:
    """Container object for Squid ACL lookup request"""
    _channel = -1
    _client = None
    _acl = list()

    def __init__(self, line):
        """Parse a lookup request from Squid into its constituent parts"""
        parts = line.decode().replace('\n', '').split(' ')

        # See if we're using concurrency; if so the first token is the integer channel ID
        try:
            self._channel = int(parts[0])
            parts.pop(0)
        except ValueError:
            pass

        # First non-channel argument must be the client IP address
        # Failure to parse the client address is handled later on
        # by detecting the object's client property being None.
        addr = parts.pop(0)
        if addr != '-':
            try:
                self._client = ipaddress.ip_address(addr)
            except ValueError:
                pass

        # Everything else is ACL arguments
        self._acl = parts

    @property
    def client(self):
        """IP address of the client that made the current request"""
        return self._client

    @property
    def acl(self):
        """List of ACL entries to test the current request against"""
        return self._acl

    def make_response(self, result='BH', pairs=dict()):
        """Create a response to this request
            result: ACL result (OK, ERR, or BH)
            pairs: dictionary of keywords to append to the response.

            See the Squid documentation for valid keywords:
            http://wiki.squid-cache.org/Features/AddonHelpers#Access_Control_.28ACL.29
        """
        chan = None
        pair = None
        keywords = {}

        # Check for valid keywords; underscore suffix is reserved for admin use
        # reference: http://wiki.squid-cache.org/Features/AddonHelpers#Access_Control_.28ACL.29 
        for key, value in pairs.items():
            if key[-1] == '_' or key in _squid_keywords:
                keywords[key] = value
            else:
                # FIXME - log this
                pass

        # Include channe if it was specified in the request
        if self._channel != -1:
            chan = str(self._channel)
        
        # Join together valid keywords
        if len(keywords) > 0:
            pair = ' '.join(['{}={}'.format(item[0], quote(item[1])) for item in keywords.items()])

        # Only include defined items in response line
        line = ' '.join([p for p in [chan, result, pair] if p is not None])+'\n\n'
        return line.encode()



import asyncio
import aioredis
import pickle
import sys

pool = None;
lock = asyncio.Lock()

# Redis key prefixes
KEY_IP = __name__ + '^ip-to-id^'
KEY_ID = __name__ + '^instance^'

# Note - this script will not work with clustered Redis due to
# use of dynamic key names. Should be fine as long as we're only
# useing a single local node.
KEY_SCRIPT = """
local instance = redis.call('get', ARGV[1]..ARGV[2])
if instance then
  return redis.call('get', ARGV[3]..instance)
end
"""

@asyncio.coroutine
def lookup(config, request):
    global pool
    global lock

    if request.client is None:
        return None

    instance_data = None

    # standard check/lock/check pattern to ensure only one thread creates a connection pool
    if pool is None:
        with (yield from lock):
            if pool is None:
                pool = yield from aioredis.create_pool((config.redis_host, config.redis_port), minsize=2, maxsize=20)

    # Call the eval script to lookup IP and retrieve instance data.
    # Could probably optimize this by storing the script server-side 
    # during initial pool creation.
    with (yield from pool) as redis:
        pickle_data = yield from redis.eval(KEY_SCRIPT, args=[KEY_IP, str(request.client), KEY_ID])
        if pickle_data is not None:
            instance_data = pickle.loads(pickle_data)

    return instance_data


@asyncio.coroutine
def store(config, instance_data):
    redis = yield from aioredis.create_redis((config.redis_host, config.redis_port))
    pipe = redis.pipeline()
    instance_id = instance_data['instance_id']
    
    # Store pickled data keyed off instance ID
    pipe.setex(KEY_ID + instance_id, float(config.redis_ttl), pickle.dumps(instance_data))

    # Store intermediate key lookups so that we can find an instance given only its IP address
    for interface in instance_data.get('network_interfaces', []):
        for address in interface.get('private_ip_addresses', []):
            pipe.setex(KEY_IP + address['private_ip_address'], float(config.redis_ttl), instance_id)
            if 'association' in interface:
                pipe.setex(KEY_IP + interface['association']['public_ip'], float(config.redis_ttl), instance_id)

    yield from pipe.execute()
    redis.close()
    yield from redis.wait_closed()

@asyncio.coroutine
def close():
    global pool
    global lock

    if pool is not None:
        with (yield from lock):
            if pool is not None:
                # Wait for all pool connections to become free, indicating that no tasks are currently using it
                while pool.freesize != pool.size:
                    yield from asyncio.sleep(1)

                # Ask the connection pool to close any open connections, and wait for it to do so
                pool.close()
                yield from pool.wait_closed()
                pool = None



# -*- coding: utf-8 -*-

import os
import sys
import click
import asyncio
import concurrent.futures

from asyncio.streams import StreamWriter, FlowControlMixin

from . import squid
from . import metadata
from . import aclmatch
from . import config

reader, writer = None, None


@asyncio.coroutine
def stdio(loop=None):
    """Set up stdin/stdout stream handlers"""
    if loop is None:
        loop = asyncio.get_event_loop()

    reader = asyncio.StreamReader()
    reader_protocol = asyncio.StreamReaderProtocol(reader)

    writer_transport, writer_protocol = yield from loop.connect_write_pipe(FlowControlMixin, os.fdopen(0, 'wb'))
    writer = StreamWriter(writer_transport, writer_protocol, None, loop)

    yield from loop.connect_read_pipe(lambda: reader_protocol, sys.stdin)

    return reader, writer


@asyncio.coroutine
def async_input(config,):
    """Handle reading lines from stdin and handing off to background task for processing"""
    loop = asyncio.get_event_loop()

    global reader, writer
    if (reader, writer) == (None, None):
        reader, writer = yield from stdio()

    while True:
        line = yield from reader.readline()
        if config.debug_enabled:
            print('STDIN: {}'.format(line), file=sys.stderr, flush=True)

        # Readline returns empty bystes string when the socket is closed
        if line == b'':
            yield from metadata.close()
            return
        
        # Process line in background task
        loop.create_task(handle_line(config, line))


@asyncio.coroutine
def handle_line(config, line):
    """Run an ACL lookup request line from Squid through the processing pipeline."""
    global writer

    request = None
    result = 'BH'
    pairs = {}
    try:
        # Get a Request object with parsed fields
        request = squid.Request(line)

        # Get metadata from Redis back-end
        hostinfo = yield from metadata.lookup(config, request)

        # Use metadata to make access decision (OK, ERR, or BH)
        result, pairs = yield from aclmatch.test(request, hostinfo)
        
    except Exception as e:
        pairs = { 'log': 'Exception encountered handling request: '+str(e) }
        # Only discard the request if we failed to parse the input from Squid - ensures
        # that errors are reported properly when using concurrency.
        if request is None:
            request = squid.Request(b'- -')
        
    # Output response to Squid
    response = request.make_response(result, pairs)
    if config.debug_enabled:
        print('STDOUT: {}'.format(response), file=sys.stderr, flush=True)

    writer.write(response)
    yield from writer.drain()

@click.option(
    '--debug',
    is_flag=True,
    help="Enable debug logging to STDERR."
)
@click.option(
    '--port',
    default=6379,
    type=int,
    help='Redis server port.'
)
@click.option(
    '--host',
    default='localhost',
    type=str,
    help='Redis server hostname.'
)
@click.command()
def listen(**args):
    """Listen for ACL lookup request lines on stdin and write the responses on stdout"""
    _config = config.Config(**args)
    loop = asyncio.get_event_loop()
    loop.set_debug(_config.debug_enabled)
    loop.run_until_complete(async_input(_config))



import click
import boto3
import asyncio

from . import metadata
from . import config

def camel_dict_to_snake_dict(camel_dict):
    """Convert Boto3 CamelCase dict to snake_case dict"""
    def camel_to_snake(name):

        import re

        first_cap_re = re.compile('(.)([A-Z][a-z]+)')
        all_cap_re = re.compile('([a-z0-9])([A-Z])')
        s1 = first_cap_re.sub(r'\1_\2', name)

        return all_cap_re.sub(r'\1_\2', s1).lower()


    def value_is_list(camel_list):

        checked_list = []
        for item in camel_list:
            if isinstance(item, dict):
                checked_list.append(camel_dict_to_snake_dict(item))
            elif isinstance(item, list):
                checked_list.append(value_is_list(item))
            else:
                checked_list.append(item)

        return checked_list


    snake_dict = {}
    for k, v in camel_dict.items():
        if isinstance(v, dict):
            snake_dict[camel_to_snake(k)] = camel_dict_to_snake_dict(v)
        elif isinstance(v, list):
            snake_dict[camel_to_snake(k)] = value_is_list(v)
        else:
            snake_dict[camel_to_snake(k)] = v

    return snake_dict


def tag_list_to_dict(tags_list):
    """Convert Boto3-style key-value tags list into dict"""
    tags_dict = {}

    for tag in tags_list:
        if 'key' in tag:
            tags_dict[tag['key']] = tag['value']
        elif 'Key' in tag:
            tags_dict[tag['Key']] = tag['Value']

    return tags_dict


def store_aws_metadata(config):
    """Store AWS metadata (result of ec2.describe_instances call) into Redis"""
    loop = asyncio.get_event_loop()
    session = boto3.Session(profile_name=config.profile_name, region_name=config.region_name)
    client = session.client('ec2')
    response = client.describe_instances()
    tasks = []

    # Find all instances, convert to snake dict, convert to tags, and fire off
    # task to store in Redis
    for reservation in response.get('Reservations', []):
        for instance in reservation.get('Instances', []):
            instance = camel_dict_to_snake_dict(instance)
            instance['tags'] = tag_list_to_dict(instance.get('tags', []))
            print('Storing data for {instance_id}'.format(**instance))
            tasks.append(loop.create_task(metadata.store(config,instance)))

    if len(tasks) > 0: 
        loop.run_until_complete(asyncio.wait(tasks))
        loop.stop()

@click.option(
    '--ttl', 
    default=1800,
    type=int, 
    help='Time-to-live for AWS metadata stored in Redis.')
@click.option(
    '--port',
    default=6379,
    type=int,
    help='Redis server port.'
)
@click.option(
    '--host',
    default='localhost',
    type=str,
    help='Redis server hostname.'
)
@click.option(
    '--region',
    default=None,
    type=str,
    help='AWS Region name (overrides region from profile).'
)
@click.option(
    '--profile',
    default=None,
    type=str,
    help='AWS Configuration Profile name.'
)
@click.command()
def sync(**args):
    """Collect inventory from EC2 and persist to Redis"""
    _config = config.Config(**args)
    store_aws_metadata(_config)





class Config:
    """Configuration object to store command-line options or defaults"""
    _redis_host = 'localhost'
    _redis_port = 6379
    _redis_ttl = 1800
    _profile_name = None
    _region_name = None
    _debug = False

    def __init__(self, host=None, port=None, ttl=None, profile=None, region=None, debug=False):
        if host is not None:
            self._redis_host = host
        if port is not None:
            self._redis_port = port
        if ttl is not None:
            self._redis_ttl = ttl
        if profile is not None:
            self._profile_name = profile
        if region is not None:
            self._region_name = region
        if debug is not False:
            self._debug = True

    @property
    def redis_host(self):
        """Hostname or address of Redis server"""
        return self._redis_host

    @property
    def redis_port(self):
        """Port of Redis server"""
        return self._redis_port

    @property
    def redis_ttl(self):
        """Expiration time for AWS metadata stored in Redis"""
        return self._redis_ttl

    @property
    def profile_name(self):
        """AWS Configuration Profile name"""
        return self._profile_name

    @property
    def region_name(self):
        """AWS Region name"""
        return self._region_name

    @property
    def debug_enabled(self):
        """Debug Flag Status"""
        return self._debug



# -*- coding: utf-8 -*-

import sys
from setuptools import setup, find_packages

if sys.version_info < (3, 4):
    raise RuntimeError("aws-acl-helper doesn't support Python versions below 3.4")

version = {}

with open("aws_acl_helper/version.py") as fp:
    exec(fp.read(), version)

with open('README.rst') as f:
    readme = f.read()

with open('LICENSE') as f:
    license = f.read()

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setup(
    name='aws-acl-helper',
    version=version['__version__'],
    description='Squid external ACL helper that allows use of AWS instance metadata',
    long_description=readme,
    author='Brandon Davidson',
    author_email='brad@oatmail.org',
    url='https://github.com/brandond/aws-acl-helper',
    download_url='https://github.com/brandond/aws-acl-helper/tarball/{}'.format(version['__version__']),
    license=license,
    packages=find_packages(exclude=('docs')),
    entry_points={
        'console_scripts': ['aws-acl-helper=aws_acl_helper.commands:cli']
    },
    include_package_data=True,
    install_requires=requirements,
)

